'#################################################
'# 
'# fbList: Fast arbitrary index insert/remove list
'# 
'# Version: 0.2
'# 
'# Official site: www.JatteGames.com
'# 
'#################################################
'# 
'# fbList is Copyright (c) 2023 Johannes "Jattenalle" Pihl, all rights reserved
'# 
'# 	This software is provided 'as-is', without any express or implied
'# 	warranty. In no event will the authors be held liable for any damages
'# 	arising from the use of this software.
'# 	
'# 	Permission is granted to anyone to use this software for any purpose,
'# 	including commercial applications, and to alter it and redistribute it
'# 	freely, subject to the following restrictions:
'# 	
'# 	1. The origin of this software must not be misrepresented; you must not
'# 	   claim that you wrote the original software. If you use this software
'# 	   in a product, an acknowledgment in the product documentation would be
'# 	   appreciated but is not required.
'# 	2. Altered source versions must be plainly marked as such, and must not be
'# 	   misrepresented as being the original software.
'# 	3. This notice may not be removed or altered from any source distribution.
'# 
'#################################################
'# 
'# Known issues:
'#   List operations are not atomic, some care must be taken if accessing the
'#   same list from multiple different threads at the same time. 
'# 
'#################################################
'# 
'# Custom sort function structure:
'#   function(byval A as any ptr, byval B as any ptr) as integer
'# 
'#################################################
'# 
'# Use #DEFINE FBLIST_INCLUDE_HEADER_ONLY before you include fbList to only
'# include the declares and definitions without the methods and procedures
'# themselves.
'# Example:
'#   #DEFINE FBLIST_INCLUDE_HEADER_ONLY
'#   #INCLUDE ONCE "fbList.bi"
'# 
'#################################################

#ifdef FBLIST_INC
	#error  __FILE__ : FBLIST is already included!
#endif
#define FBLIST_INC

#define FBLIST_VERSION			"0.2"

#include once "crt/string.bi"

const LIST_FIRST	=	0
const LIST_LAST		=	-1


type t_List
	declare sub Alloc(byval entries as integer)
	
	declare function Shift(byval _first as integer, byval _entries as integer, byval toIndex as integer = LIST_LAST) as integer
	
	declare function Fill(byval _p as integer, byval _flen as integer, byval index as integer = LIST_LAST) as integer
	declare function Fill(byval _p as any ptr, byval _flen as integer, byval index as integer = LIST_LAST) as integer
	
	declare function Insert(byval _p as integer, byval index as integer = LIST_LAST) as integer
	declare function Insert(byval _p as any ptr, byval index as integer = LIST_LAST) as integer
	declare function Insert(_v() as integer, byval index as integer = LIST_LAST) as integer
	declare function Insert(_v() as any ptr, byval index as integer = LIST_LAST) as integer
	declare function Insert(byref _l as t_List, byval index as integer = LIST_LAST) as integer
	
	declare function SortedInsert(byval _p as integer) as integer
	declare function SortedInsert(byval _p as any ptr) as integer
	declare function SortedInsert(_v() as integer) as integer
	declare function SortedInsert(_v() as any ptr) as integer
	declare function SortedInsert(byref _l as t_List) as integer
	
	declare function Replace(byval _p as integer, byval index as integer) as integer
	declare function Replace(byval _p as any ptr, byval index as integer) as integer
	declare function Replace(_v() as integer, byval index as integer) as integer
	declare function Replace(_v() as any ptr, byval index as integer) as integer
	declare function Replace(byref _l as t_List, byval index as integer) as integer
	
	declare function Merge(byval _p as integer, byval index as integer) as integer
	declare function Merge(byval _p as any ptr, byval index as integer) as integer
	declare function Merge(_v() as integer, byval index as integer) as integer
	declare function Merge(_v() as any ptr, byval index as integer) as integer
	declare function Merge(byref _l as t_List, byval index as integer) as integer
	
	declare sub Clear()
	declare sub Clear(byval _p as integer)
	declare sub Clear(byval _p as any ptr)
	declare sub Clear(_v() as integer)
	declare sub Clear(_v() as any ptr)
	declare sub Clear(byref _l as t_List)
	
	
	declare function Find(byval _match as integer, byval _matchIndex as integer = 0) as integer
	declare function Find(byval _match as any ptr, byval _matchIndex as integer = 0) as integer
	
	declare sub Remove(byval index as integer)
	
	declare sub Switch(byval a as integer, byval b as integer)
	
	declare function GetInt(byval index as integer) as integer
	declare function GetPtr(byval index as integer) as any ptr
	
	declare function CSV() as string
	
	declare operator [] (byval index as integer) as any ptr
	
	declare sub EnableSort()
	declare sub DisableSort()
	
	declare function IsSorted() as integer
	
	declare sub AllowDuplicates()
	declare sub DisallowDuplicates()
	
	declare sub Sort()
	declare sub Shuffle(byval depth as integer = 1)
	declare sub RemoveDuplicates()
	
	declare sub DefaultCompareFunction()
	declare sub CustomCompareFunction(byval pfnCompare as any ptr)
	declare sub DefaultDeleteSub()
	declare sub CustomDeleteSub(byval pfnDelete as any ptr)
	
	declare Constructor ()
	declare Constructor(byref _p as integer)
	declare Constructor(byref _p as any ptr)
	declare Constructor(_v() as integer)
	declare Constructor(_v() as any ptr)
	declare Constructor(byref _l as t_List)
	
	declare operator let(byref _p as integer)
	declare operator let(byref _p as any ptr)
	declare operator let(_v() as integer)
	declare operator let(_v() as any ptr)
	declare operator let(byref _l as t_List)
	
	declare operator +=(byref _p as integer)
	declare operator +=(byref _p as any ptr)
	#if 0 'FreeBASIC doesn't actually support this!
		declare operator +=(_v() as integer)
		declare operator +=(_v() as any ptr)
	#endif
	declare operator +=(byref _l as t_List)
	
	declare Destructor ()
	
	declare function Length() as integer
	declare function SizeOf() as integer
	declare function AllocSize() as integer
	
	RawListMemory as any ptr ptr
	
	declare sub __ForEachTarget()
	#if __FB_VERSION__ >= "1.08"
		'# Just adds support for optional parenthesis, syntax: list.ForEach(i as integer) OR list.ForEach i as integer
		#macro ForEach?(_a)
			__ForEachTarget()
			for _a = 0 to ___T_LIST_FOREACHTARGET->Length - 1
		#endmacro
	#else
		'# Older FBC versions do not support optional parenthesis, syntax: list.ForEach(i as integer)
		#macro ForEach(_a)
			__ForEachTarget()
			for _a = 0 to ___T_LIST_FOREACHTARGET->Length - 1
		#endmacro
	#endif
	
	private:
		declare function __Insert(byref _p as any ptr, byval index as integer, byval _sortparam as integer) as integer
		declare function __FitIndex(byref index as integer) as integer
		declare sub __GrowAlloc()
		
		__pfnDelete as sub(byval _a as any ptr)
		__pfnCompare as function(byval _a as any ptr, byval _b as any ptr) as integer
		
		_AllowDuplicate as integer	=	TRUE
		_ForceSorted as integer		=	FALSE
		_IsSorted as integer		=	TRUE
		_Length as integer
		
		_AllocSize as integer
end type
dim shared as t_List ptr ___T_LIST_FOREACHTARGET

'#################################################
'# 
'# Detect possible shenanigans and throw an error
'# If this occurs for you then something is very wrong with your compiler optimization parameters
'# Removing this check to get rid of the error while keeping the weird compiler parameters will probably result in crashes
'# 
'#################################################
dim shared as integer ___T_LIST_ENTRYSIZE
___T_LIST_ENTRYSIZE = sizeof(any ptr)
if sizeof(integer) <> ___T_LIST_ENTRYSIZE then
	dim as integer f = freefile
	open cons for output as #f
	print #f, "ERROR: "& __FILE__ &"("& __LINE__ &") - sizeof INTEGER ("& sizeof(integer) &") is different than sizeof ANY PTR ("& sizeof(any ptr) &")"
	print #f, "	This not normal behaviour! Some compiler memory optimizations are destructive!"
	close #f
	error -17
	end
end if


#ifndef FBLIST_INCLUDE_HEADER_ONLY
	declare function ___T_LIST_DEFCMP(byval _a as any ptr, byval _b as any ptr) as integer
	function ___T_LIST_DEFCMP(byval _a as any ptr, byval _b as any ptr) as integer
		if cast(integer, _a) < cast(integer, _b) then
			return -1
		elseif cast(integer, _a) > cast(integer, _b) then
			return 1
		else
			return 0
		end if
	end function
	Constructor t_List ()
		this.__pfnCompare = @___T_LIST_DEFCMP
	end constructor
	Constructor t_List(byref _p as integer)
		this.constructor()
		this.insert(_p)
	end constructor
	Constructor t_List(byref _p as any ptr)
		this.constructor()
		this.insert(_p)
	end constructor
	Constructor t_List(_v() as integer)
		this.constructor()
		this.insert(_v())
	end constructor
	Constructor t_List(_v() as any ptr)
		this.constructor()
		this.insert(_v())
	end constructor
	Constructor t_List(byref _l as t_List)
		this.constructor()
		this.insert(_l)
	end constructor
	
	Destructor t_List ()
		if this.__pfnDelete then
			for g as integer = 0 to this._Length - 1
				this.__pfnDelete(this.RawListMemory[g])
			next
		end if
		delete This.RawListMemory
		This.RawListMemory		=	0
		This._Length	=	0
		This._AllocSize	=	0
	end Destructor
	
	operator t_List.let(byref _p as integer)
		this.Clear()
		this.Insert(_p)
	end operator
	operator t_List.let(byref _p as any ptr)
		this.Clear()
		this.Insert(_p)
	end operator
	operator t_List.let(_v() as integer)
		this.Clear()
		this.Insert(_v())
	end operator
	operator t_List.let(_v() as any ptr)
		this.Clear()
		this.Insert(_v())
	end operator
	operator t_List.let(byref _l as t_List)
		this.Clear()
		this.Insert(_l)
	end operator
	
	operator t_List.+=(byref _p as integer)
		this.Insert(_p)
	end operator
	operator t_List.+=(byref _p as any ptr)
		this.Insert(_p)
	end operator
	#if 0 'FreeBASIC doesn't actually support this!
		operator t_List.+=(_v() as integer)
			this.Insert(_v())
		end operator
		operator t_List.+=(_v() as any ptr)
			this.Insert(_v())
		end operator
	#endif
	operator t_List.+=(byref _l as t_List)
		this.Insert(_l)
	end operator
	
	operator t_List.[] (byval index as integer) as any ptr
		if index = LIST_LAST then
			index = This._Length - 1
		end if
		if index < 0 or index >= This._Length then
			return 0
		end if
		return this.RawListMemory[index]
	end operator
	
	function t_List.__FitIndex(byref index as integer) as integer
		if index = LIST_LAST then
			index = this._Length
			return 1
		end if
		if index > this._Length then
			index = this._Length
			return 1
		end if
		if index < 0 then
			index = 0
			return -1
		end if
		return 0
	end function
	
	
	
	
	
	
	sub t_List.__ForEachTarget()
		___T_LIST_FOREACHTARGET = @this
	end sub
	sub t_List.EnableSort()
		this._ForceSorted = TRUE
	end sub
	sub t_List.DisableSort()
		this._ForceSorted = FALSE
	end sub
	
	sub t_List.AllowDuplicates()
		this._AllowDuplicate = TRUE
	end sub
	sub t_List.DisallowDuplicates()
		this._AllowDuplicate = FALSE
	end sub
	
	sub t_List.DefaultCompareFunction()
		this.__pfnCompare = @___T_LIST_DEFCMP
	end sub
	sub t_List.CustomCompareFunction(byval pfnCompare as any ptr)
		this.__pfnCompare = pfnCompare
	end sub
	
	sub t_List.DefaultDeleteSub()
		this.__pfnDelete = 0
	end sub
	sub t_List.CustomDeleteSub(byval pfnDelete as any ptr)
		this.__pfnDelete = pfnDelete
	end sub
	
	function t_List.IsSorted() as integer
		return this._IsSorted
	end function
	
	function t_List.Length() as integer
		return this._Length
	end function
	function t_List.SizeOf() as integer
		return this._Length * ___T_LIST_ENTRYSIZE
	end function
	function t_List.AllocSize() as integer
		return this._AllocSize
	end function
	
	
	
	
	
	
	function t_List.GetPtr(byval index as integer) as any ptr
		if index = LIST_LAST then
			index = This._Length - 1
		end if
		if index < 0 or index >= This._Length then
			return 0
		end if
		return cptr(any ptr, This.RawListMemory[index])
	end function
	
	function t_List.GetInt(byval index as integer) as integer
		if index = LIST_LAST then
			index = This._Length - 1
		end if
		if index < 0 or index >= This._Length then
			return 0
		end if
		return cast(integer, This.RawListMemory[index])
	end function
	
	
	
	
	
	
	function t_List.CSV() as string
		dim as string r
		if this._Length then
			for i as integer = 0 to this._Length - 1
				r &= cint(this[i])
				if i < this._Length -1 then
					r &=  ", "
				end if
			next
		end if
		return r
	end function
	
	
	
	
	
	
	sub t_List.Switch(byval a as integer, byval b as integer)
		if a = LIST_LAST then
			a = This._Length - 1
		end if
		if a < 0 or a >= This._Length then
			exit sub
		end if
		if b = LIST_LAST then
			b = This._Length - 1
		end if
		if b < 0 or b >= This._Length then
			exit sub
		end if
		if a <> b then
			swap This.RawListMemory[a], This.RawListMemory[b]
			this._IsSorted		=	FALSE
		end if
	end sub
	
	
	
	
	
	
	sub t_List.Clear()
		if This._Length then
			if this.__pfnDelete then
				for g as integer = 0 to this._Length - 1
					this.__pfnDelete(this.RawListMemory[g])
				next
			end if
			memset(This.RawListMemory, 0, This._Length * ___T_LIST_ENTRYSIZE)
			This._Length		=	0
		end if
	end sub
	sub t_List.Clear(byval _p as integer)
		this.Clear()
		this.Insert(_p)
	end sub
	sub t_List.Clear(byval _p as any ptr)
		this.Clear()
		this.Insert(_p)
	end sub
	sub t_List.Clear(_v() as integer)
		this.Clear()
		this.Insert(_v())
	end sub
	sub t_List.Clear(_v() as any ptr)
		this.Clear()
		this.Insert(_v())
	end sub
	sub t_List.Clear(byref _l as t_List)
		this.Clear()
		this.Insert(_l)
	end sub
	
	
	
	
	
	
	sub t_List.Alloc(byval entries as integer)
		if entries > this._AllocSize then
			dim as integer addSize = int(this._Length * 0.1) '// Add 10% of the current list size
			if addSize > 1024 then '// Don't add more than this extra, even if we could
				addSize = 1024
			end if
			entries += addSize
			
			dim as integer mSize = entries * ___T_LIST_ENTRYSIZE
			this._AllocSize = entries
			
			if this._Length = 0 then
				deallocate this.RawListMemory
				this.RawListMemory = callocate(mSize)
				this._AllocSize = entries
			else
				dim as any ptr _hold = This.RawListMemory
				This.RawListMemory		=	callocate(mSize)
				memcpy(This.RawListMemory, _hold, this._Length * ___T_LIST_ENTRYSIZE)
				deallocate _hold
				
			end if
		end if
	end sub
	
	
	
	
	
	function t_List.Shift(byval _first as integer, byval _entries as integer, byval toIndex as integer = LIST_LAST) as integer
		this.__FitIndex(_first)
		
		if _entries = LIST_LAST then
			_entries = this._Length - _first
		end if
		
		if toIndex = LIST_LAST then
			toIndex = this._Length - _entries
		end if
		this.__FitIndex(toIndex)
		
		if _entries > 0 then
			if toIndex + _entries <= this._Length then
				if _first + _entries < this._Length then
					dim as any ptr _hold = callocate(_entries * ___T_LIST_ENTRYSIZE)
					if _hold = 0 then error 12
					memcpy(_hold, This.RawListMemory + _first, _entries * ___T_LIST_ENTRYSIZE)
					memmove(This.RawListMemory + _first, This.RawListMemory + (_first + _entries), (this._Length - _first - _entries) * ___T_LIST_ENTRYSIZE)
					memmove(This.RawListMemory + toIndex + _entries, This.RawListMemory + toIndex, (this._Length - toIndex -_entries) * ___T_LIST_ENTRYSIZE)
					memcpy(This.RawListMemory + toIndex, _hold, _entries * ___T_LIST_ENTRYSIZE)
					deallocate _hold
				end if
			end if
		end if
		return toIndex
	end function
	
	
	
	
	
	sub t_List.Remove(byval index as integer)
		if index = LIST_LAST then
			index = This._Length - 1
		end if
		if index < 0 or index >= This._Length then
			exit sub
		end if
		if this.__pfnDelete then
			this.__pfnDelete(this.RawListMemory[index])
		end if
		memmove(This.RawListMemory + index, This.RawListMemory + index +1, (This._Length-index-1) * ___T_LIST_ENTRYSIZE)
		This.RawListMemory[This._Length - 1] = 0
		This._Length				-=	1
	end sub
	
	
	
	
	
	
	function t_List.Fill(byval _p as integer, byval _flen as integer, byval index as integer = LIST_LAST) as integer
		return This.Fill(cptr(any ptr, _p), _flen, index)
	end function
	function t_List.Fill(byval _p as any ptr, byval _flen as integer, byval index as integer = LIST_LAST) as integer
		if index = LIST_LAST then
			index = This._Length
		end if
		if index < 0 or index >= This._Length then
			return -1
		end if
		if _flen < 1 then
			return -1
		end if
		
		this.Alloc(This._Length + _flen)
		
		dim as any ptr _hold = This.RawListMemory
		This.RawListMemory		=	callocate(This._AllocSize * ___T_LIST_ENTRYSIZE)
		if index > 0 then
			memcpy(This.RawListMemory, _hold, index * ___T_LIST_ENTRYSIZE)
		end if
		if index < This._Length then
			memcpy(This.RawListMemory + (index + _flen), _hold + (index * ___T_LIST_ENTRYSIZE), (This._Length-index) * ___T_LIST_ENTRYSIZE)
		end if
		deallocate _hold
		for f as integer = 0 to _flen - 1
			This.RawListMemory[index + f]	=	_p
		next
		
		This._Length			+=	_flen
		this._IsSorted		=	FALSE
		
		return index
	end function
	
	
	
	
	
	
	function t_List.Merge(byref _l as t_list, byval index as integer) as integer
		this.__FitIndex(index)
		this.Alloc(This._Length + _l._Length)
		for g as integer = 0 to _l._Length
			index = this.Merge(_l.GetPtr(g), index) + 1
		next
		return index
	end function
	function t_List.Merge(_v() as any ptr, byval index as integer) as integer
		this.__FitIndex(index)
		dim as integer _lb = lbound(_v)
		dim as integer _l = ubound(_v) - _lb
		this.Alloc(This._Length + _l)
		for g as integer = 0 to _l
			index = this.Merge(_v(_lb + g), index) + 1
		next
		return index
	end function
	function t_List.Merge(_v() as integer, byval index as integer) as integer
		this.__FitIndex(index)
		dim as integer _lb = lbound(_v)
		dim as integer _l = ubound(_v) - _lb
		this.Alloc(This._Length + _l)
		for g as integer = 0 to _l
			index = this.Merge(_v(_lb + g), index) + 1
		next
		return index
	end function
	function t_List.Merge(byval _p as integer, byval index as integer) as integer
		return this.Merge(cptr(any ptr, _p), index)
	end function
	function t_List.Merge(byval _p as any ptr, byval index as integer) as integer
		this.__FitIndex(index)
		if this.__pfnDelete then
			if this.RawListMemory[index] then
				this.__pfnDelete(this.RawListMemory[index])
			end if
		end if
		if index = this._Length then
			index = this.Insert(_p)
		else
			This.RawListMemory[index]	=	_p
		end if
		this._IsSorted		=	FALSE
		return index
	end function
	
	
	
	
	
	
	function t_List.Replace(byref _l as t_List, byval index as integer) as integer
		this.__FitIndex(index)
		for g as integer = 0 to _l._Length
			index = this.Replace(_l.GetPtr(g), index) + 1
		next
		return index
	end function
	function t_List.Replace(_v() as any ptr, byval index as integer) as integer
		this.__FitIndex(index)
		dim as integer _lb = lbound(_v)
		dim as integer _l = ubound(_v) - _lb
		for g as integer = 0 to _l
			index = this.Replace(_v(_lb + g), index) + 1
		next
		return index
	end function
	function t_List.Replace(_v() as integer, byval index as integer) as integer
		this.__FitIndex(index)
		dim as integer _lb = lbound(_v)
		dim as integer _l = ubound(_v) - _lb
		for g as integer = 0 to _l
			index = this.Replace(_v(_lb + g), index) + 1
		next
		return index
	end function
	function t_List.Replace(byval _p as integer, byval index as integer) as integer
		return this.Replace(cptr(any ptr, _p), index)
	end function
	function t_List.Replace(byval _p as any ptr, byval index as integer) as integer
		this.__FitIndex(index)
		if index < this._Length then
			if this.__pfnDelete then
				if this.RawListMemory[index] then
					this.__pfnDelete(this.RawListMemory[index])
				end if
			end if
			This.RawListMemory[index]	=	_p
			this._IsSorted		=	FALSE
		end if
		return index
	end function
	
	
	
	
	
	
	function t_List.__Insert(byref _p as any ptr, byval index as integer, byval _sortparam as integer) as integer
		this.__FitIndex(index)
		
		this._IsSorted		=	_sortparam
		
		this.Alloc(This._Length + 1)
		
		if this._AllowDuplicate = FALSE then
			dim as integer r = this.Find(_p)
			if r > -1 then
				return r
			end if
		end if
		
		if index < This._Length then '// Anywhere except adding to the end
			if This._Length + 1 > This._AllocSize then '// Grow it, and split it while we're at it
				This._AllocSize	+=	1
			end if
			dim as any ptr _hold = This.RawListMemory
			This.RawListMemory		=	callocate(This._AllocSize * ___T_LIST_ENTRYSIZE)
			if index > 0 then
				memcpy(This.RawListMemory, _hold, index * ___T_LIST_ENTRYSIZE)
			end if
			memcpy(This.RawListMemory + (index + 1), _hold + (index * ___T_LIST_ENTRYSIZE), (This._Length-index) * ___T_LIST_ENTRYSIZE)
			deallocate _hold
			
			This.RawListMemory[index]	=	_p
			This._Length			+=	1
			
			return index
		else
			This.RawListMemory[This._Length]	=	_p
			This._Length				+=	1
			return This._Length - 1
		end if
		return -1
	end function
	
	
	
	
	
	
	function t_List.Insert(byref _l as t_List, byval index as integer = LIST_LAST) as integer
		this.__FitIndex(index)
		if @this = @_l then '// Special case for self
			if this._AllowDuplicate then
				this.Alloc(this._Length * 2) '// Prepare for worst case
				
				memcpy(this.RawListMemory + this._Length, this.RawListMemory, this._Length * ___T_LIST_ENTRYSIZE)
				this._Length *= 2
				this._IsSorted = FALSE
				
				if this._ForceSorted then
					this.Sort()
				end if
			end if
		else
			for g as integer = 0 to _l._Length - 1
				index = this.insert(_l.GetPtr(g), index) + 1
			next
		end if
		return index
	end function
	function t_List.Insert(_v() as integer, byval index as integer = LIST_LAST) as integer
		this.__FitIndex(index)
		dim as integer _lb = lbound(_v)
		dim as integer _l = ubound(_v) - _lb
		for g as integer = 0 to _l
			index = this.insert(_v(_lb + g), index) + 1
		next
		return index
	end function
	function t_List.Insert(_v() as any ptr, byval index as integer = LIST_LAST) as integer
		this.__FitIndex(index)
		dim as integer _lb = lbound(_v)
		dim as integer _l = ubound(_v) - _lb
		for g as integer = 0 to _l
			index = this.insert(_v(_lb + g), index) + 1
		next
		return index
	end function
	function t_List.Insert(byval _p as integer, byval index as integer = LIST_LAST) as integer
		if this._ForceSorted then
			return This.SortedInsert(cptr(any ptr, _p))
		end if
		return This.__Insert(cptr(any ptr, _p), index, FALSE)
	end function
	function t_List.Insert(byval _p as any ptr, byval index as integer = LIST_LAST) as integer
		if this._ForceSorted then
			return This.SortedInsert(_p)
		end if
		return this.__Insert(_p, index, FALSE)
	end function
	
	
	
	
	
	
	function t_List.SortedInsert(byref _l as t_List) as integer
		dim as integer oldOption = this._ForceSorted
		dim as integer r
		this._ForceSorted = TRUE
		r = this.Insert(_l)
		this._ForceSorted = oldOption
		return r
	end function
	function t_List.SortedInsert(_v() as integer) as integer
		dim as integer oldOption = this._ForceSorted
		dim as integer r
		this._ForceSorted = TRUE
		r = this.Insert(_v())
		this._ForceSorted = oldOption
		return r
	end function
	function t_List.SortedInsert(_v() as any ptr) as integer
		dim as integer oldOption = this._ForceSorted
		dim as integer r
		this._ForceSorted = TRUE
		r = this.Insert(_v())
		this._ForceSorted = oldOption
		return r
	end function
	function t_List.SortedInsert(byval _p as integer) as integer
		return This.SortedInsert(cptr(any ptr, _p))
	end function
	function t_List.SortedInsert(byval _p as any ptr) as integer
		if this._IsSorted = FALSE then '// Sort the existing list
			this.Sort()
		end if
		
		if this._Length then
			dim as integer _length	=	This._Length - 1
			dim as integer pivot	=	_length / 2
			dim as integer min		=	0
			dim as integer max		=	_length
			
			dim as integer compareResult
			
			do while _length > 0
				compareResult = this.__pfnCompare(_p, This.RawListMemory[pivot])
				if compareResult < 0 then '// Less
					max		=	pivot - 1
					_length	=	max - min
					pivot	=	min + _length / 2
				elseif compareResult > 0 then
					min		=	pivot + 1
					_length	=	max - min
					pivot	=	min + _length / 2
				else '// Duplicate
					if this._AllowDuplicate then
						return this.__Insert(_p, pivot, TRUE)
					end if
					return pivot
				end if
			loop
			compareResult = this.__pfnCompare(_p, This.RawListMemory[pivot])
			if compareResult > 0 then
				pivot += 1
			elseif compareResult = 0 then '// Duplicate
				if this._AllowDuplicate then
					return this.__Insert(_p, pivot, TRUE)
				end if
				return pivot
			end if
			if pivot < 0 then
				pivot = 0
			end if
			
			return this.__Insert(_p, pivot, TRUE)
		else '// Just insert it, there's nothing else to do
			return this.__Insert(_p, LIST_LAST, TRUE)
		end if
		return -1
	end function
	
	
	
	
	
	
	sub t_List.Sort()
		this._IsSorted		=	TRUE
		if this._Length < 2 then '// There's either no entries, or just one entry, in the list so we don't need to sort it
			exit sub
		else
			dim as integer compareResult
			for p as integer = 0 to this._Length - 2
				for c as integer = p+1 to this._Length - 1
					compareResult = this.__pfnCompare(This.RawListMemory[p], This.RawListMemory[c])
					if compareResult = 0 then '// Duplicate
						
					elseif compareResult > 0 then
						this.Switch(p, c)
					end if
				next
			next
		end if
	end sub
	
	
	
	
	
	
	sub t_List.Shuffle(byval depth as integer = 1)
		dim as integer r
		for d as integer = 1 to depth
			for i as integer = 0 to this._Length - 1
				r = cint(this._Length * rnd)
				this.Switch(i, r)
			next
		next
	end sub
	
	
	
	
	
	
	function t_List.Find(byval _match as integer, byval _matchIndex as integer = 0) as integer
		return this.Find(cptr(any ptr, _match), _matchIndex)
	end function
	function t_List.Find(byval _match as any ptr, byval _matchIndex as integer = 0) as integer
		if this._Length then
			dim as integer _matchCount
			dim as integer compareResult
			if this._IsSorted then
				dim as integer _length	=	This._Length - 1
				dim as integer pivot	=	_length / 2
				dim as integer min		=	0
				dim as integer max		=	_length
				
				do while _length > 0
					if This.RawListMemory[pivot] = _match then
						if _matchCount = _matchIndex then
							return pivot
						end if
						_matchCount += 1
					else
						compareResult = this.__pfnCompare(_match, This.RawListMemory[pivot])
						if compareResult < 0 then '// Less
							max		=	pivot - 1
							_length	=	max - min
							pivot	=	min + _length / 2
						elseif compareResult > 0 then
							min		=	pivot + 1
							_length	=	max - min
							pivot	=	min + _length / 2
						else '// Duplicate
							if _matchCount = _matchIndex then
								return pivot
							end if
							_matchCount += 1
						end if
					end if
				loop
				if This.RawListMemory[pivot] = _match then
					if _matchCount = _matchIndex then
						return pivot
					end if
					_matchCount += 1
				end if
				compareResult = this.__pfnCompare(_match, This.RawListMemory[pivot])
				if compareResult = 0 then '// Duplicate
					if _matchCount = _matchIndex then
						return pivot
					end if
					_matchCount += 1
				end if
			else
				for p as integer = 0 to this._Length - 1
					if This.RawListMemory[p] = _match then
						if _matchCount = _matchIndex then
							return p
						end if
						_matchCount += 1
					else
						compareResult = this.__pfnCompare(This.RawListMemory[p], _match)
						if compareResult = 0 then
							if _matchCount = _matchIndex then
								return p
							end if
							_matchCount += 1
						end if
					end if
				next
			end if
		end if
		return -1
	end function
	
	
	
	
	
	
	sub t_List.RemoveDuplicates()
		dim as integer compareResult
		if this._IsSorted then
			for p as integer = 0 to this._Length - 2
				if p >= this._Length then
					exit for
				end if
				for c as integer = p + 1 to this._Length - 1
					if c >= this._Length then
						exit for
					end if
					if This.RawListMemory[p] = This.RawListMemory[c] then
						this.Remove(c)
						c -= 1
					else
						compareResult = this.__pfnCompare(This.RawListMemory[p], This.RawListMemory[c])
						if compareResult = 0 then
							this.Remove(c)
							c -= 1
						else
							exit for '// The list is sorted, so if the values don't match then there can be no more duplicates
						end if
					end if
				next
			next
		else
			for p as integer = 0 to this._Length - 1
				if p >= this._Length then
					exit for
				end if
				for c as integer = p+1 to this._Length - 1
					if c >= this._Length then
						exit for
					end if
					if This.RawListMemory[p] = This.RawListMemory[c] then
						this.Remove(c)
						c -= 1
					else
						compareResult = this.__pfnCompare(This.RawListMemory[p], This.RawListMemory[c])
						if compareResult = 0 then
							this.Remove(c)
							c -= 1
						end if
					end if
				next
			next
		end if
	end sub
#endif
